module BagWithHeap (Bag, empty, insert, mkBag, isEmpty, union, minBag, delMin) where

import Graphviz 

type Bag a   = HTree a 
data HTree a = Null | Fork Int a (HTree a) (HTree a) deriving (Show)

-- Definition der Gleichheit von Bags
instance Ord a => Eq (HTree a)
  where
  b1 == b2 = qsort (flatten b1) == qsort (flatten b2)

flatten :: Bag a -> [a]
flatten Null = []
flatten (Fork p i lt rt) = flatten lt ++ [i] ++ flatten rt

qsort    :: Ord a => [a] -> [a]
qsort []     = []
qsort (x:xs) = qsort [y | y <- xs, y < x] ++ [x] ++ qsort [y | y <- xs, y >= x]


-- Methoden

fork     :: a -> HTree a -> HTree a -> HTree a
fork x yt zt = if m < n then Fork p x zt yt else Fork p x yt zt
  where m = size yt
        n = size zt
        p = m + n + 1

size     :: HTree a -> Int
size Null             = 0
size (Fork n x yt zt) = n

empty    :: Bag a
empty = Null

insert   :: (Ord a) => a -> Bag a -> Bag a
insert x bx = union (mkBag [x]) bx

mkBag    :: (Ord a) => [a] -> Bag a
mkBag xs = fst (mkTwo (length xs) xs)

mkTwo    :: (Ord a) => Int -> [a] -> (Bag a, [a])
mkTwo n xs
  | (n == 0) = (Null, xs)
  | (n == 1) = (fork (head xs) Null Null, tail xs)
  | otherwise = (union xt yt, zs)
  where (xt, ys) = mkTwo m xs
        (yt, zs) = mkTwo (n - m) ys
        m        = n `div` 2

isEmpty  :: Bag a -> Bool
isEmpty Null = True
isEmpty (Fork n x yt zt) = False

union    :: (Ord a) => Bag a -> Bag a -> Bag a
union Null yt = yt
union (Fork m u vt wt) Null = Fork m u vt wt
union (Fork m u vt wt) (Fork n x yt zt)
  | (u <= x)  = fork u vt (union wt (Fork n x yt zt))
  | otherwise = fork x yt (union (Fork m u vt wt) zt)

minBag   :: Bag a -> a
minBag (Fork n x yt zt) = x

delMin   :: (Ord a) => Bag a -> Bag a
delMin (Fork n x yt zt) = union yt zt


-- Visualisierung des Heaps

exampleBag1 = insert 1 (insert 3 (insert 2 (insert 6 (insert 4 empty))))
exampleBag2 = insert 2 (insert 2 (insert 1 (insert 1 (insert 3 empty))))
exampleBag3 = mkBag [12,23,12,77,33,24,23,53]


toDot :: Show a => Bag a -> String
toDot s =
  "digraph tmp {\n" ++
  concat nodes ++
  concat edges ++
  "}\n"
  where
  (nodes, edges, cnt, id) = getNodes [] [] 0 s
  getNodes :: (Show a) => [String] -> [String] -> Int -> (Bag a) -> ([String], [String], Int, String)
  getNodes nodes edges cnt Null = (nodes, edges, cnt, "")
  getNodes nodes edges cnt (Fork i x Null Null) = 
    (("  node [label=\"Node " ++ dataString ++ "\",shape=box]; " ++ idString ++ ";\n") : nodes, edges, cnt+1, idString)
    where
    idString = show cnt
    dataString = "d: " ++ show x ++ " / i: " ++ show i
  getNodes nodes edges cnt (Fork i x lt rt) = 
    (("  node [label=\"Node " ++ dataString ++ "\",shape=ellipse]; " ++ idString ++ ";\n") : nodesLeft,
     edgeLeft : edgeRight : edgesLeft,
     cntLeft, idString)
    where
    (nodesLeft,  edgesLeft,  cntLeft,  idLeft)  = getNodes nodesRight edgesRight (cntRight) lt
    (nodesRight, edgesRight, cntRight, idRight) = getNodes nodes      edges      (cnt+1)    rt
    idString  = show cnt
    edgeLeft  = if idLeft  /= "" then "  " ++ idString ++ " -> " ++ idLeft  ++ ";\n" else ""
    edgeRight = if idRight /= "" then "  " ++ idString ++ " -> " ++ idRight ++ ";\n" else ""
    dataString = "d: " ++ show x ++ " / i: " ++ show i

debugDisplay :: Show a => Bag a -> String
debugDisplay s = toDot s

display :: Show a => Bag a -> IO ()
display s =
  do
  displayGraph (toDot s)
