module FlexArray (empty, size, isEmpty, access, update, hiext, hirem, loext, lorem) where
import Graphviz

data Flex a = Null | Leaf a | Fork Int (Flex a) (Flex a)

empty :: Flex a
empty = Null

isEmpty :: Flex a -> Bool
isEmpty Null           = True
isEmpty (Leaf a)       = False
isEmpty (Fork n xs yt) = False

size :: Flex a -> Int
size Null           = 0
size (Leaf x)       = 1
size (Fork n xt yt) = n

height :: Flex a -> Int
height Null = 0
height (Leaf a) = 0
height (Fork n xt yt) = 1 + (height xt `max` height yt)

access :: Flex a -> Int -> a
access (Leaf x) 0 = x
access (Fork n xt yt) k =
	if k < m then access xt k
	         else access yt (k-m)
	where m = size xt

update :: Flex a -> Int -> a -> Flex a
update (Leaf y) 0 x = Leaf x
update (Fork n xt yt) k x
	 | k < m     = Fork n (update xt k x) yt
	 | otherwise = Fork n xt (update yt (k-m) x)
	 where m = size xt

hiext1 :: a -> Flex a -> Flex a
hiext1 x Null           = Leaf x
hiext1 x (Leaf y)       = Fork 2 (Leaf y) (Leaf x)
hiext1 x (Fork n xt yt) = Fork (n+1) (Fork n xt yt) (Leaf x)

hiext2 :: a -> Flex a -> Flex a
hiext2 x Null = Leaf x
hiext2 x (Leaf y) = Fork 2 (Leaf y) (Leaf x)
hiext2 x (Fork n xt yt)
	| size xt == size yt = Fork (n+1) (Fork n xt yt) (Leaf x)
	| otherwise          = Fork (n+1) xt (hiext2 x yt)

hiext :: a -> Flex a -> Flex a
hiext x Null = Leaf x
hiext x (Leaf y) = Fork 2 (Leaf y) (Leaf x)
hiext x (Fork n xt yt) = Fork (n+1) xt (hiext2 x yt)

hirem :: Flex a -> Flex a
hirem (Leaf x) = Null
hirem (Fork n xt yt) = fork (n-1) xt (hirem yt)

fork :: Int -> Flex a -> Flex a -> Flex a
fork n xt yt
	| isEmpty xt = yt
	| isEmpty yt = xt
	| otherwise  = Fork n xt yt

loext :: a -> Flex a -> Flex a
loext x Null = Leaf x
loext x (Leaf y) = Fork 2 (Leaf x) (Leaf y)
loext x (Fork n xt yt) = Fork (n+1) (lo x xt) yt
	where
		lo x (Leaf y) = Fork 2 (Leaf x) (Leaf y)
		lo x (Fork n xt yt)
			| size xt == size yt = Fork (n+1) (Leaf x) (Fork n xt yt)
			| otherwise          = Fork (n+1) (lo x xt) yt

lorem :: Flex a -> Flex a
lorem (Leaf x) = Null
lorem (Fork n xt yt) = fork (n-1) (lorem xt) yt

toList :: Flex a -> [a]
toList Null           = []
toList (Leaf y)       = [y]
toList (Fork x xt yt) = toList xt ++ toList yt

--fromList :: [a] -> Flex a
--fromList []     = empty
--fromList (x:xs) = 

instance Show a => Show (Flex a) where
	show x = show (toList x)

toDot :: Show a => Flex a -> String
toDot x = 
	"digraph tmp {\n" ++
	concat nodes ++
	concat edges ++
	"}\n"
	where
		(nodes, edges, cnt, id) = hlp [] [] 0 x
		hlp :: Show a => [String] -> [String] -> Int -> (Flex a) -> ([String], [String], Int, String)
		hlp nodes edges cnt Null = (
			("  node [label=\"Null\",shape=box]; "++idStr++";\n"):nodes,
			edges,
			cnt,
			""
			)
			where idStr = show cnt
		hlp nodes edges cnt (Leaf x) = (
			("  node [label=\"Node "++(show x)++"\",shape=box]; "++idStr++";\n"):nodes,
			edges,
			cnt+1,
			idStr
			)
			where idStr = show cnt
		hlp nodes edges cnt (Fork x xt yt) = (
			("  node [label=\"Fork "++(show x)++"\",shape=ellipse]; "++idStr++";\n"):nodes2,
			("  "++idStr++" -> "++id2++";\n") : ("  "++idStr++" -> "++id1++";\n") : edges2,
			cnt2+1,
			idStr
			) 
			where
				(nodes1, edges1, cnt1, id1) = hlp nodes edges cnt yt
				(nodes2, edges2, cnt2, id2) = hlp nodes1 edges1 cnt1 xt
				idStr = show cnt2

display :: Show a => Flex a -> IO ()
display x =
	do
	displayGraph (toDot x)

