module SetWithTree (Set, empty, isEmpty, member, insert, delete, union, meet, minus) where

import Graphviz

type Set a   = STree a
data STree a = Null | Fork (STree a) a (STree a) deriving (Show)

-- Definition der Gleichheit von Mengen
instance Ord a => Eq (STree a)
  where
  s1 == s2 = qsort (flatten s1) == qsort(flatten s2)
  
flatten :: Set a -> [a]
flatten Null = []
flatten (Fork lt i rt) = flatten lt ++ [i] ++ flatten rt

qsort    :: Ord a => [a] -> [a]
qsort []     = []
qsort (x:xs) = qsort [y | y <- xs, y < x] ++ [x] ++ qsort [y | y <- xs, y >= x]


--Methoden

empty    :: Set a
empty = Null

isEmpty  :: Set a -> Bool
isEmpty Null           = True
isEmpty (Fork xt y zt) = False

member   :: Ord a => Set a -> a -> Bool
member Null x = False
member (Fork xt y zt) x
  | (x <  y)  = member xt x
  | (x == y)  = True
  | otherwise = member zt x

insert   :: Ord a => a -> Set a -> Set a
insert x Null = Fork Null x Null
insert x (Fork xt y zt)
  | (x <  y)  = Fork(insert x xt) y zt
  | (x == y)  = Fork xt y zt
  | otherwise = Fork xt y (insert x zt)

delete   :: Ord a => a -> Set a -> Set a
delete x Null = Null
delete x (Fork xt y zt)
  | (x <  y)  = Fork(delete x xt) y zt
  | (x == y)  = join xt zt
  | otherwise = Fork xt y (delete x zt)

join     :: Set a -> Set a -> Set a
join xt yt = if isEmpty yt then xt else Fork xt y zt
  where (y, zt) = splitTree yt

splitTree :: Set a -> (a, Set a)
splitTree (Fork xt y zt) = if isEmpty xt then (y, zt) else (u, Fork vt y zt)
  where (u, vt) = splitTree xt

union    :: (Ord a) => Set a -> Set a -> Set a
union s Null = s
union Null s = s
union s (Fork xt y zt) = union (union (insert y s) xt) zt

meet     :: (Ord a) => Set a -> Set a -> Set a
meet s Null = Null
meet Null s = Null
meet s (Fork xt y zt) = if member s y then insert y (union sl sr) else union sl sr
  where
  sl = meet s xt
  sr = meet s zt

minus    :: (Ord a) => Set a -> Set a -> Set a
minus s Null = s
minus Null s = Null
minus (Fork xt y zt) s = if member s y then union sl sr else insert y (union sl sr)
  where
  sl = minus xt s
  sr = minus zt s


-- Visualisierung der Baumes

exampleTree    = insert 1 (insert 3 (insert 2 (insert 6 (insert 4 empty))))
littleTree     = insert 5 (insert 7 empty)
unbalancedTree = insert 6 (insert 4 (insert 3 (insert 2 (insert 1 empty))))
fullTree       = insert 7 (insert 5 (insert 1 (insert 3 (insert 2 (insert 6 (insert 4 empty))))))

toList :: Set a -> [a]
toList Null           = []
toList (Fork lt x rt) = toList lt ++ [x] ++ toList rt

--instance Show a => Show (STree a) where
--	show x = show (toList x)

toDot :: Show a => Set a -> String
toDot s =
  "digraph tmp {\n" ++
  concat nodes ++
  concat edges ++
  "}\n"
  where
  (nodes, edges, cnt, id) = getNodes [] [] 0 s
  getNodes :: (Show a) => [String] -> [String] -> Int -> (Set a) -> ([String], [String], Int, String)
  getNodes nodes edges cnt Null = (nodes, edges, cnt, "")
  getNodes nodes edges cnt (Fork Null x Null) = 
    (("  node [label=\"Node " ++ dataString ++ "\",shape=box]; " ++ idString ++ ";\n") : nodes, edges, cnt+1, idString)
    where
    idString = show cnt
    dataString = show x
  getNodes nodes edges cnt (Fork lt x rt) = 
    (("  node [label=\"Node " ++ dataString ++ "\",shape=ellipse]; " ++ idString ++ ";\n") : nodesLeft,
     edgeLeft : edgeRight : edgesLeft,
     cntLeft, idString)
    where
    (nodesLeft,  edgesLeft,  cntLeft,  idLeft)  = getNodes nodesRight edgesRight (cntRight) lt
    (nodesRight, edgesRight, cntRight, idRight) = getNodes nodes      edges      (cnt+1)    rt
    idString  = show cnt
    edgeLeft  = if idLeft  /= "" then "  " ++ idString ++ " -> " ++ idLeft  ++ ";\n" else ""
    edgeRight = if idRight /= "" then "  " ++ idString ++ " -> " ++ idRight ++ ";\n" else ""
    dataString = show x

debugDisplay :: Show a => Set a -> String
debugDisplay s = toDot s

display :: Show a => Set a -> IO ()
display s =
  do
  displayGraph (toDot s)
